/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "ScoreCreationInfo.h"

ScoreCreationInfo::ScoreCreationInfo (FiledataPtr f) : track_included(f->getNumTracks()), track_node(f->getNumTracks()), track_name(f->getNumTracks())
{
	fd = f;
	division = f->getDivision();
	int i;
	for (i = 0; i < f->getNumTracks(); i++)
		track_included[i] = False;

	calculate_track_names();

	num_msgs = 0;
	num_msgs_allocated = 30;
	msgs = (MIDI_msg**) malloc(sizeof(MIDI_msg*) * num_msgs_allocated);
	if (msgs == NULL)
		ew.e("malloc failed");

	num_funcs = 0;
	num_funcs_allocated = 30;
	funcs = (FuncEl**) malloc(sizeof(FuncEl*) * num_funcs_allocated);
	if (funcs == NULL)
		ew.e("malloc failed");
}

ScoreCreationInfo::ScoreCreationInfo (uint2 div) : track_included(0), track_node(0), track_name(0)
{
	fd = NULL;
	division = div;

	num_msgs = 0;
	num_msgs_allocated = 30;
	msgs = (MIDI_msg**) malloc(sizeof(MIDI_msg*) * num_msgs_allocated);
	if (msgs == NULL)
		ew.e("malloc failed");

	num_funcs = 0;
	num_funcs_allocated = 30;
	funcs = (FuncEl**) malloc(sizeof(FuncEl*) * num_funcs_allocated);
	if (funcs == NULL)
		ew.e("malloc failed");
}

ScoreCreationInfo::~ScoreCreationInfo ()
{
	int i;
	for (i = 0; i < num_msgs; i++)
		delete msgs[i];
	free(msgs);
	for (i = 0; i < num_funcs; i++)
		delete funcs[i];
	free(funcs);
}

err_ind ScoreCreationInfo::addTrack(const char* t, ioref n)
{
	int i;
	for (i = 0; i < track_name.length; i++)
		if (*(track_name[i]) == t)
			break;
	if (i == track_name.length)
		return err; // track doesn't exist
	if (track_included[i])
		return no_err; // track already added
	track_included[i] = True;
	track_node[i] = n;
	return no_err;
}

int ScoreCreationInfo::addRemainingTracks(ioref n)
{
	int result = 0;
	int i;
	for (i = 0; i < track_name.length; i++)
		if (!track_included[i])
		{
			track_included[i] = True;
			track_node[i] = n;
			result++;
		}
	return result;
}

int ScoreCreationInfo::addRemainingTracksAndLookupNodenames()
{
	int result = 0;
	int i;
	for (i = 0; i < track_name.length; i++)
		if (!track_included[i])
		{
			ioref n = oap->open_output(*(track_name[i]));
			track_included[i] = True;
			track_node[i] = n;
			result++;
		}
	return result;
}

void ScoreCreationInfo::addMIDI_msg(const MIDI_msg& m)
{
	if (num_msgs == num_msgs_allocated)
	{
		num_msgs_allocated *= 2;
		msgs = (MIDI_msg**) realloc(msgs, sizeof(MIDI_msg*) * num_msgs_allocated);
		if (msgs == NULL)
			ew.e("realloc failed");
	}
	msgs[num_msgs++] = new MIDI_msg(m);
}

void ScoreCreationInfo::addFunction(void (*func)(void*, ScorePlayer*), void* v, ScheduleTime t)
{
	if (num_funcs == num_funcs_allocated)
	{
		num_funcs_allocated *= 2;
		funcs = (FuncEl**) realloc(funcs, sizeof(FuncEl*) * num_funcs_allocated);
		if (funcs == NULL)
			ew.e("realloc failed");
	}
	FuncEl* tmp = new FuncEl;
	tmp->func = func;
	tmp->v = v;
	tmp->time = t;
	funcs[num_funcs++] = tmp;
}

void ScoreCreationInfo::calculate_track_names()
{
	int i;
	for (i = 0; i < track_name.length; i++)
	{
		Filedata::MidiMsgArrayPtr p1 = fd->getTrack(i);
		int j;
		for (j = 0; j < p1->length; j++) // find track name
		{
			MidiMsg* m = &((*p1)[j]);
			if (m->is_ext() && m->ed->st == smsgt::file_meta &&
				 m->ed->meta_event.mt == meta::txt_seq_or_trk)
			{
				track_name[i] = (m->ed->meta_event.txt);
				break;
			}
		}
		if (track_name[i] == NULL)
		{
			char foo[20];
			sprintf(foo,"TRACK_%d",i);
			track_name[i] = new UtilString(foo);
		}
	}
}

void ScoreCreationInfo::sort (MIDI_msg** els, int len)
{
	if (len < 2)
		return;
	Bool go_again = True;
	while (go_again == True)
	{
		go_again = False;
		for (int i = 0; i < len - 1; i++)
			if (els[i]->time > els[i+1]->time)
			{
				MIDI_msg* temp = els[i];
				els[i] = els[i+1];
				els[i+1] = temp;
				go_again = True;
			}
	}
}

void ScoreCreationInfo::sort (FuncEl** els, int len)
{
	if (len < 2)
		return;
	Bool go_again = True;
	while (go_again == True)
	{
		go_again = False;
		for (int i = 0; i < len - 1; i++)
			if (els[i]->time > els[i+1]->time)
			{
				FuncEl* temp = els[i];
				els[i] = els[i+1];
				els[i+1] = temp;
				go_again = True;
			}
	}
}

// Author: Patrick Pelletier
